@@ -2,63 +2,181 @@
 Implements the JVC protocol
 """
 
-import datetime
 import logging
-from typing import Union
+from typing import Final, Union, Awaitable, Callable
 import asyncio
-
-# import the enums
+from jvc_projector.optimal_settings import hdr_film, hdr_game, sdr_film, sdr_game
 from jvc_projector.commands import ACKs, Footer, Header, Commands, PowerStates, Enum
 
 
 class JVCProjector:
     """JVC Projector Control"""
 
-    # Const values
-    PJ_OK = ACKs.greeting.value
-    PJ_ACK = ACKs.pj_ack.value
-    req = ACKs.pj_req.value
-
     def __init__(
         self,
         host: str,
-        port: int = 20554,
-        delay_ms: int = 1000,
-        connect_timeout: int = 10,
-        password: str = None,
+        password: str = "",
+        # Can supply a logger object. It can hook into the HA logger
         logger: logging.Logger = logging.getLogger(__name__),
+        connection_lost_callback: Callable[[], Awaitable[None]] = None,
+        loop: asyncio.AbstractEventLoop = None,
+        update_callback: Callable[[str], None] = None,
+        port: int = 20554,
+        connect_timeout: int = 10,
     ):
         self.host = host
         self.port = port
-        self.connect_timeout = connect_timeout
-        self.delay = datetime.timedelta(microseconds=(delay_ms * 1000))
-        self.last_command_time = datetime.datetime.now() - datetime.timedelta(
-            seconds=10
-        )
         # NZ models have password authentication
         self.password = password
+        self.connect_timeout: int = connect_timeout
         self.logger = logger
+        self._update_callback = update_callback
+        # use the provided loop or get current one. Otherwise make one
+        try:
+            self._loop = loop or asyncio.get_running_loop()
+        except RuntimeError:
+            self._loop = asyncio.new_event_loop()
         self._lock = asyncio.Lock()
+        # Const values
+        self.PJ_OK: Final = ACKs.greeting.value
+        self.PJ_ACK: Final = ACKs.pj_ack.value
+        self.PJ_REQ: Final = ACKs.pj_req.value
+        self._connection_lost_callback = connection_lost_callback
+        self._closed = False
+        self._retry_interval = 1
+        self._closing = False
+        self._halted = False
+        self.reader: asyncio.StreamReader = None
+        self.writer: asyncio.StreamWriter = None
 
+    async def async_open_connection(self) -> bool:
+        """Open a connection"""
+        assert self.port >= 0, f"Port must be greater than 0: {self.port}"
+        assert self.host != "", f"Host must not be empty: {self.host}"
+        msg, success = await self.reconnect()
+        if not success:
+            self.logger.error(msg)
+
+        return success
+
+    async def reconnect(self):
+        """Initiate keep-alive connection"""
+        while True:
+            try:
+                if self._halted:
+                    await asyncio.sleep(2)
+
+                self.logger.debug(
+                    "Connecting to JVC Projector: %s:%s", self.host, self.port
+                )
+                async with self._lock:
+                    # transport, protocol = await self._loop.create_connection(
+                    #     lambda: self.protocol, self.host, self.port
+                    # )
+                    self.reader, self.writer = await asyncio.open_connection(
+                        self.host, self.port, loop=self._loop
+                    )
+                    self.logger.debug("Connected to JVC Projector")
+                    # create a reader and writer to do handshake
+                async with self._lock:
+                    self.logger.debug("Handshaking")
+                    result, success = await self._async_handshake()
+                    if not success:
+                        return result, success
+                    self.logger.debug("Handshake complete and we are connected")
+                self._reset_retry_interval()
+                return "Connection done", True
+
+            # includes conn refused
+            except OSError as err:
+                self._increase_retry_interval()
+                interval = self._get_retry_interval()
+                self.logger.warning(
+                    "Connecting failed, retrying in %i seconds", interval
+                )
+                if not self._closing:
+                    return f"Connection failed: {err}", False
+                await asyncio.sleep(interval)
+            except asyncio.TimeoutError:
+                self._increase_retry_interval()
+                interval = self._get_retry_interval()
+                self.logger.warning(
+                    "Connection timed out, retrying in %i seconds", interval
+                )
+                if not self._closing:
+                    return "Connection timed out", False
+                await asyncio.sleep(interval)
+
+    async def _async_handshake(self) -> tuple[str, bool]:
+        """
+        Do the 3 way handshake
+
+        Projector sends PJ_OK, client sends PJREQ (with optional password) within 5 seconds, projector replies with PJACK
+        first, after connecting, see if we receive PJ_OK. If not, raise exception
+        """
         if self.password:
-            self.pj_req = self.req + f"_{self.password}".encode()
+            pj_req = self.PJ_REQ + f"_{self.password}".encode()
         else:
-            self.pj_req = self.req
+            pj_req = self.PJ_REQ
 
-    async def _async_throttle(self):
-        if self.delay == 0:
-            return
+        # 3 step handshake
+        msg_pjok = await self.reader.read(len(self.PJ_OK))
+        self.logger.debug(msg_pjok)
+        if msg_pjok != self.PJ_OK:
+            result = f"Projector did not reply with correct PJ_OK greeting: {msg_pjok}"
+            success = False
 
-        delta = datetime.datetime.now() - self.last_command_time
+            return result, success
 
-        if self.delay > delta:
-            return await asyncio.sleep((self.delay - delta).total_seconds())
+        # try sending PJREQ, if there's an error, raise exception
+        try:
+            self.writer.write(pj_req)
+            await self.writer.drain()
+        except asyncio.TimeoutError as err:
+            return f"Timeout sending PJREQ {err}", False
+
+        # see if we receive PJACK, if not, raise exception
+        msg_pjack = await self.reader.read(len(self.PJ_ACK))
+        if msg_pjack != self.PJ_ACK:
+            return f"Exception with PJACK: {msg_pjack}", False
+
+        return "ok", True
+
+    async def connection_lost(self):
+        """restart connection"""
+        if not self._closing:
+            await self.reconnect()
+
+    def _get_retry_interval(self):
+        return self._retry_interval
+
+    def _reset_retry_interval(self):
+        self._retry_interval = 1
+
+    def _increase_retry_interval(self):
+        self._retry_interval = min(300, 1.5 * self._retry_interval)
+
+    def close(self):
+        # TODO: maybe implement these
+        """Close the AVR device connection and don't try to reconnect."""
+        self.logger.debug("Closing connection to AVR")
+        self._closing = True
+
+    def halt(self):
+        """Close the AVR device connection and wait for a resume() request."""
+        self.logger.warning("Halting connection to AVR")
+        self._halted = True
+
+    def resume(self):
+        """Resume the AVR device connection if we have been halted."""
+        self.logger.warning("Resuming connection to AVR")
+        self._halted = False
 
     async def _async_send_command(
         self,
         send_command: Union[list[bytes], bytes],
-        ack: bytes,
         command_type: bytes = b"!",
+        ack: bytes = None,
     ) -> tuple[str, bool]:
         """
         Sends a command with a flag to expect an ack.
@@ -77,92 +195,50 @@ class JVCProjector:
             )
         """
 
-        result = ""
-        success = True
+        # Check commands
+        if command_type == Header.reference.value:
+            result, success = await self._async_do_command(
+                send_command, ack, command_type
+            )
 
-        # throttle command if too quick otherwise
-        # JVC kills the connection
-        await self._async_throttle()
-
-        async with self._lock:
-            # Try connecting
-            try:
-                reader, writer = await asyncio.open_connection(self.host, self.port)
-            except ConnectionRefusedError:
-                return "Connection Refused", False
-            except asyncio.TimeoutError:
-                return "Connection timed out", False
-
-        # 3 step handshake
-        result, success = await self._async_handshake(reader, writer)
-        if not success:
             return result, success
 
-        # Check commands
         if isinstance(send_command, list):
+
             for cmd in send_command:
+                cons_command, ack = await self._async_construct_command(
+                    cmd, command_type
+                )
+                if not ack:
+                    return cons_command, ack
+                # need a delay otherwise it kills connection
+                await asyncio.sleep(0.1)
                 result, success = await self._async_do_command(
-                    reader, writer, cmd, ack, command_type
+                    cons_command, ack.value, command_type
                 )
                 if not success:
                     return result, success
         else:
+            try:
+                cons_command, ack = await self._async_construct_command(
+                    send_command, command_type
+                )
+            except TypeError:
+                cons_command = send_command
+
+            if not ack:
+                return cons_command, ack
             result, success = await self._async_do_command(
-                reader, writer, send_command, ack, command_type
+                cons_command, ack.value, command_type
             )
             if not success:
                 return result, success
 
-        self.last_command_time = datetime.datetime.now()
         self.logger.debug("send command result: %s", result)
-
-        return result, success
-
-    async def _async_handshake(
-        self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter
-    ) -> tuple[str, bool]:
-        """
-        Do the 3 way handshake
-
-        Projector sends PJ_OK, client sends PJREQ within 5 seconds, projector replies with PJACK
-        first, after connecting, see if we receive PJ_OK. If not, raise exception
-        """
-
-        async with self._lock:
-            msg_pjok = await reader.read(len(self.PJ_OK))
-            if msg_pjok != self.PJ_OK:
-                result = (
-                    f"Projector did not reply with correct PJ_OK greeting: {msg_pjok}"
-                )
-                success = False
-
-                return result, success
-
-            # try sending PJREQ, if there's an error, raise exception
-            try:
-                writer.write(self.pj_req)
-                await writer.drain()
-            except asyncio.TimeoutError as err:
-                result = f"Timeout sending PJREQ {err}"
-                success = False
-                writer.close()
-
-                return result, success
-
-            # see if we receive PJACK, if not, raise exception
-            msg_pjack = await reader.read(len(self.PJ_ACK))
-            if msg_pjack != self.PJ_ACK:
-                result = f"Exception with PJACK: {msg_pjack}"
-                success = False
-
-                return result, success
-
-            return "ok", True
+        return "ok", True
 
     async def _async_do_command(
         self,
-        reader: asyncio.StreamReader,
-        writer: asyncio.StreamWriter,
         command: bytes,
         ack: bytes,
         command_type: bytes = b"!",
@@ -170,11 +246,19 @@ class JVCProjector:
         async with self._lock:
             self.logger.debug("_do_command sending command: %s", command)
             # send the command
-            writer.write(command)
-            await writer.drain()
+            self.writer.write(command)
+            try:
+                await self.writer.drain()
+            except ConnectionError as err:
+                # reaching this means the writer was closed somewhere
+                self.logger.error(err)
+                # restart the connection
+                await self.connection_lost()
+                self.writer.write(command)
+                await self.writer.drain()
 
             # if we send a command that returns info, the projector will send
-            # an ack, followed by the message. Check to see if the ack sent by
+            # an ack, followed by the actual message. Check to see if the ack sent by
             # projector is correct, then return the message.
             ack_value = (
                 Header.ack.value + Header.pj_unit.value + ack + Footer.close.value
@@ -183,11 +267,11 @@ class JVCProjector:
 
             # Receive the acknowledgement from PJ
             try:
-                received_ack = await reader.read(len(ack_value))
+                received_ack = await self.reader.read(len(ack_value))
             except asyncio.TimeoutError:
+                # Sometimes if you send a command that is greyed out, the PJ will just hang
                 result = "Connection timed out. Command is probably not allowed to run at this time."
                 self.logger.error(result)
-                writer.close()
                 return result, False
             except ConnectionRefusedError:
                 self.logger.error("Connection Refused when getting ack")
@@ -195,24 +279,21 @@ class JVCProjector:
 
             self.logger.debug("received_ack: %s", received_ack)
 
+            # get the ack for operation
+            if received_ack == ack_value and command_type == Header.operation.value:
+                self.logger.debug("result: %s", received_ack)
+                return received_ack, True
+
             # if we got what we expect and this is a reference,
             # receive the data we requested
-            if received_ack == ack_value and command_type == Header.operation.value:
-                result = received_ack
-                writer.close()
-
-                return result, True
-                
             if received_ack == ack_value and command_type == Header.reference.value:
-                message = await reader.read(1024)
+                message = await self.reader.read(1024)
                 self.logger.debug("result: %s, %s", received_ack, message)
-                writer.close()
-                result = message
 
-                return result, True
-            
+                return message, True
+
             # Otherwise, it failed
-            writer.close()
+            # Because this now reuses a connection, reaching this stage means catastrophic failure, or HA running as usual :)
             result = "Unexpected ack received from PJ after sending a command. Perhaps a command got cancelled because a new connection was made."
             self.logger.error(result)
             self.logger.error("received_ack: %s", received_ack)
@@ -224,9 +305,13 @@ class JVCProjector:
         self, raw_command: str, command_type: bytes
     ) -> tuple[bytes, ACKs]:
         """
-        Transform commands into their byte values
+        Transform commands into their byte values from the string value
         """
-        command, value = raw_command.split(",")
+        # split command into the base and the action like menu: left
+        try:
+            command, value = raw_command.split(",")
+        except ValueError:
+            return "No value for command provided", False
 
         # Check if command is implemented
         if not hasattr(Commands, command):
@@ -248,7 +333,7 @@ class JVCProjector:
         self, command: Union[list[str], str], command_type: bytes = b"!"
     ) -> tuple[str, bool]:
         """
-        Sync version of async_exec_command
+        Sync wrapper for async_exec_command
         """
 
         return asyncio.run(self.async_exec_command(command, command_type))
@@ -271,36 +356,11 @@ class JVCProjector:
             )
         """
         self.logger.debug("exec_command Executing command: %s", command)
-
-        if isinstance(command, str):
-            cons_command, ack = await self._async_construct_command(
-                command, command_type
-            )
-            result, success = await self._async_send_command(
-                cons_command, ack.value, command_type
-            )
-
-            return result, success
-        elif isinstance(command, list):
-            for cmd in command:
-                # TODO: this may need a sleep to prevent the thing from dying
-                # TODO: also maybe implement a retry mechanism
-                cons_command, ack = await self._async_construct_command(
-                    cmd, command_type
-                )
-                result, success = await self._async_send_command(
-                    cons_command, ack.value, command_type
-                )
-
-                if not success:
-                    return result, success
-            return "ok", True
-        else:
-            return "Command is not valid type", False
+        return await self._async_send_command(command, command_type)
 
     async def async_info(self) -> tuple[str, bool]:
         """
-        Brings up the Info screen
+        Bring up the Info screen
         """
         cmd = (
             Header.operation.value
@@ -311,7 +371,7 @@ class JVCProjector:
 
         return await self._async_send_command(
             cmd,
-            ack=ACKs.menu_ack.value,
+            ack=ACKs.menu_ack,
             command_type=Header.operation.value,
         )
 
@@ -319,7 +379,7 @@ class JVCProjector:
         self,
     ) -> tuple[str, bool]:
         """
-        Turns on PJ
+        sync wrapper for async_power_on
         """
         return asyncio.run(self.async_power_on())
 
@@ -335,7 +395,7 @@ class JVCProjector:
         self,
     ) -> tuple[str, bool]:
         """
-        Turns off PJ
+        sync wrapper for async_power_off
         """
         return asyncio.run(self.async_power_off())
 
@@ -349,12 +409,14 @@ class JVCProjector:
         """
         Sets (opinionated!) optimal HDR gaming settings
         """
+        # TODO: use Userx modes for this
         state = await self.async_get_low_latency_state()
+        await asyncio.sleep(3)
         # If LL is on, we can turn it off first
         # TODO: make this more DRY
         if state:
             cmds = [
-                "picture_mode, hdr10",
+                "picture_mode, hdr",
                 "enhance, seven",
                 "motion_enhance, off",
                 "graphic_mode, hires1",
@@ -366,7 +428,7 @@ class JVCProjector:
         else:
             # If LL is off, we can enable these settings
             cmds = [
-                "picture_mode, hdr10",
+                "picture_mode, hdr",
                 "laser_dim, off",
                 "low_latency, on",
                 "enhance, seven",
@@ -379,9 +441,15 @@ class JVCProjector:
         """
         Sets (opinionated!) optimal gaming settings
         """
+        # task = asyncio.create_task(self.async_get_low_latency_state())
         state = await self.async_get_low_latency_state()
+        # todo: fix this, its a concurrency issue
+        await asyncio.sleep(3)
         # If LL is on, we can turn it off first
-        # TODO: make this more DRY
+
+        if state is None:
+            return "Error getting low latency state"
+
         if state:
             cmds = [
                 "enhance, seven",
@@ -392,7 +460,7 @@ class JVCProjector:
             # ll_commands = self._build_mode_commands(cmds)
             # return ll_commands
             return await self.async_exec_command(cmds)
-        else:
+        elif state is False:
             # If LL is off, we can enable these settings
             cmds = [
                 "laser_dim, off",
@@ -411,6 +479,7 @@ class JVCProjector:
         so PJ socket will time out
         """
         state = await self.async_get_low_latency_state()
+        await asyncio.sleep(3)
         # If LL is on, we can turn it off first
         # TODO: make this more DRY
         if state:
@@ -430,7 +499,7 @@ class JVCProjector:
             # If LL is off, we can enable these settings
             cmds = [
                 "picture_mode, frame_adapt_hdr",
-                "laser_dim, auto1",
+                "laser_dim, off",
                 "enhance, seven",
                 "motion_enhance, low",
                 "graphic_mode, hires1",
@@ -443,6 +512,7 @@ class JVCProjector:
         Sets (opinionated) optimal sdr film settings
         """
         state = await self.async_get_low_latency_state()
+        await asyncio.sleep(3)
         # If LL is on, we need to turn it off
         # TODO: make this more DRY
         if state:
@@ -508,9 +578,17 @@ class JVCProjector:
         return asyncio.run(self.async_get_low_latency_state())
 
     async def async_get_low_latency_state(self) -> bool:
-        """Get the current state of LL"""
-        state, _ = await self._async_do_reference_op("low_latency", ACKs.picture_ack)
+        """
+        Get the current state of LL
+
+        None if there was an error
+        """
+        state, success = await self._async_do_reference_op(
+            "low_latency", ACKs.picture_ack
+        )
         # LL is off, could be disabled
+        if not success:
+            return None
         if state == b"PM0":
             return False
 
